--[[
**********************************************************************
MagicDKP - a raid loot and DKP tracker.
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

local MINOR_VERSION = tonumber(("$Revision: 98 $"):match("%d+"))

MagicDKP = LibStub("AceAddon-3.0"):NewAddon("MagicDKP", "AceConsole-3.0",
					    "AceEvent-3.0", "AceTimer-3.0", "LibLogger-1.0")
MagicDKP.MAJOR_VERSION = "MagicDKP-1.0"
MagicDKP.MINOR_VERSION = MINOR_VERSION

local MagicComm = LibStub("MagicComm-1.0")
local GUI = LibStub("AceGUI-3.0")
local BabbleBoss = LibStub("LibBabble-Boss-3.0") 
local LDBIcon = LibStub("LibDBIcon-1.0", true)
local LDB = LibStub("LibDataBroker-1.1", true)
local BL = BabbleBoss:GetUnstrictLookupTable()
local GetItemCount, GetItemInfo, match = GetItemCount, GetItemInfo, string.match
BabbleBoss = nil

MagicDKPDB = {}

local mod = MagicDKP
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)

local UnitGUID = UnitGUID
local sub = string.sub
local UnitClassification = UnitClassification
local table_insert = table.insert
local loggingDeaths = false

local defaultOptions = {
   profile = {
      logLevel = mod.logLevels.INFO,
      raidsize = "25",
      trackwipe = false,
      wipepercentage = 50,
      standbyattendee = false,
      eventnameaszone = false,
      trackonly = false,
      reportraid = true,
      reportstandby = true,
      enabled = true,
      detectAlts = true,
      minimapIcon = {},
      maxraidage = 1,
   }
}

local db, zones, currentZone, currentRaid, currentEvent, currentBoss
local playerName = UnitName("player")

local networkData = {}

local function SetNetworkData(cmd, data, misc1, misc2, misc3, misc4)
   networkData.cmd   = cmd
   networkData.data  = data
   networkData.misc1 = misc1
   networkData.misc2 = misc2
   networkData.misc3 = misc3
   networkData.misc4 = misc4
end
      

-- Dummy debug ftw.
function mod:get()
   return {
      zone = currentZone,
      raid = currentRaid,
      event = currentEvent
   }
end

local function UpdateUpvalues()
   db = mod.db.profile         -- profile data
   zones = MagicDKPDB.zones     -- raid zones / boss mobs
   items = MagicDKPDB.items     -- item database
   events = MagicDKPDB.events   -- custom events
   raids = MagicDKPDB.raids     -- raids
   players = MagicDKPDB.players -- player db
end

function mod:OnInitialize()
   self.db = LibStub("AceDB-3.0"):New("MagicDKPConfigDB", defaultOptions, "Default")
   self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileDeleted","OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

   MagicDKPDB.zones    = MagicDKPDB.zones    or {}
   MagicDKPDB.items    = MagicDKPDB.items    or mod.defaultItems
   MagicDKPDB.events   = MagicDKPDB.events   or mod.defaultEvents  
   MagicDKPDB.raids    = MagicDKPDB.raids    or {}
   MagicDKPDB.players  = MagicDKPDB.players  or {}

--   mod.defaultItems  = nil 
   mod.defaultEvents = nil 
   
   UpdateUpvalues()

   self:SetLogLevel(db.logLevel)
   self:AddDefaultZones() -- Populate MagicDKPDB with the default zones
   self:BuildConfiguration()   
   if LDB then
      self.ldb =
	 LDB:NewDataObject("MagicDKP",
			   {
			      type =  "launcher", 
			      label = "MagicDKP",
			      icon = "Interface\\AddOns\\MagicDKP\\dkp.tga",
			      tooltiptext = L["|cffffffffMagicDKP|r\n|cffffff00Left click|r to toggle the configuration screen.\n"],
			      OnClick = function(clickedframe, button)
					   if button == "LeftButton" then
					      mod:ToggleConfigDialog()
					   end
					end,
			   })
      if LDBIcon then
	 LDBIcon:Register("MagicDKP", self.ldb, mod.db.profile.minimapIcon)
      end
   end
end



local currentNoteItem = {}
 noteItemQueue = {}
local itemNoteText = L["Enter value for %s, looted by %s."]
local itemLootDialog

local function CheckQueuedItem()
   local count =  #noteItemQueue
   local item = count > 0 and noteItemQueue[count]
   if item then
      mod:AddItemDialog(item.looter, item.itemid, item.value, item.link, item.count)
      noteItemQueue[count] = nil
   else
      currentNoteItem.itemid = nil
      currentNoteItem.looter = nil
      currentNoteItem.value = nil
      currentNoteItem.count = nil
      itemLootDialog:Hide()
   end
end

local function QueueItem(looter, itemid, value, link, count)
   local queueSize = #noteItemQueue
   if queueSize > 0 or currentNoteItem.itemid then
      noteItemQueue[#noteItemQueue+1] = {
	 looter = looter,
	 itemid = itemid,
	 value = value or 0,
	 link = link,
	 count = count 
      }
   else
      mod:AddItemDialog(looter, itemid, value or 0, link, count)
   end
   
end

-- Call external addons adjusting the dkp of a player
-- Currently used only when spending DKP.
function mod:AdjustExternalDKP(looter, value)
   if ShowDKPAdjustPlayerDKP then
      ShowDKPAdjustPlayerDKP(looter, value)
   end
end

function mod:AddItem(looter, itemid, value, count)
   if self.hasTrace then
      self:trace("Adding item %sx%d for %s at value %d.", items[itemid].itemInfo[2], count, looter, value)
   end
   if currentRaid then
      mod:AdjustExternalDKP(looter, value)
      local newitem = {
	 itemid = itemid,
	 looter = looter,
	 value = value,
	 count = count
      }
      if currentEvent then
	 currentEvent.items[#currentEvent.items+1] = newitem
      else
	 local key,name = self:GetZoneName(currentZone.name)
	 if not currentRaid.events[name] then
	    currentRaid.events[name] = {
	       attendees = { },
	       eventTime = self:GetGameTimeAsUnix(), -- set event time to time of first looted item
	       items = {}
	    }
	 end
	 local attendees = currentRaid.events[name].attendees
	 for i = 1, GetNumRaidMembers() do  
	    local name = GetRaidRosterInfo(i)
	    attendees[GetRaidRosterInfo(i)] = true
	 end
	 attendees[looter] = true
	 currentRaid.events[name].items[#currentRaid.events[name].items+1] = newitem
      end

      for id,data in pairs(raids) do
	 if data == currentRaid then
	    self:AddRaidConfig(id, build)
	    break;
	 end
      end
      self:NotifyChange()
   end  
end

function mod:AddItemDialog(looter, itemid, value, link, count)
   currentNoteItem.itemid = itemid
   currentNoteItem.looter = looter
   currentNoteItem.value = value
   currentNoteItem.count = count
   
   if not itemLootDialog then
      itemLootDialog = GUI:Create("MDKP_Dialog")
      itemLootDialog:SetTitle(L["Add Item"])
      itemLootDialog:SetLayout("Flow")

      itemLootDialog.ok = GUI:Create("Button")
      itemLootDialog.cancel = GUI:Create("Button")
      itemLootDialog.bank = GUI:Create("Button")
      itemLootDialog.de = GUI:Create("Button")

      itemLootDialog:AddChild(itemLootDialog.ok)
      itemLootDialog:AddChild(itemLootDialog.cancel)
      itemLootDialog:AddChild(itemLootDialog.bank)
      itemLootDialog:AddChild(itemLootDialog.de)

      itemLootDialog.ok:SetText(L["Ok"])
      itemLootDialog.cancel:SetText(L["Ignore"])
      itemLootDialog.bank:SetText(L["Bank"])
      itemLootDialog.de:SetText(L["Disenchant"])

      local onAccept = function()
			  local val = itemLootDialog:GetValue()
			  mod:AddItem(currentNoteItem.looter, currentNoteItem.itemid, val, currentNoteItem.count)
			  items[currentNoteItem.itemid].value = val
			  CheckQueuedItem()
		       end
      
      itemLootDialog:SetCallback("OnAccept", onAccept)
      itemLootDialog.ok:SetCallback("OnClick", onAccept)
      itemLootDialog.cancel:SetCallback("OnClick",
				    function()
				       items[currentNoteItem.itemid].ignore = true
				       items[currentNoteItem.itemid].value = nil
				       items[currentNoteItem.itemid].bank = nil
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.bank:SetCallback("OnClick",
				    function()
				       items[currentNoteItem.itemid].bank = true
				       items[currentNoteItem.itemid].value = nil
				       mod:AddItem("bank", currentNoteItem.itemid, 0, currentNoteItem.count)
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.de:SetCallback("OnClick",
				    function()
				       mod:AddItem("disenchant", currentNoteItem.itemid, 0, currentNoteItem.count)
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.ok:SetWidth(80)
      itemLootDialog.cancel:SetWidth(80)
      itemLootDialog.bank:SetWidth(80)
      itemLootDialog.de:SetWidth(125)
   end

   
   itemLootDialog:SetText(itemNoteText:format(link, looter ))
   itemLootDialog:SetValue(value)

   itemLootDialog:Show()
   return itemLootDialog
   
end


function mod:OnEnable()
   if not db.enabled then return end
   
   local popup = _G.StaticPopupDialogs
   if type(popup) ~= "table" then
      popup = {}
   end
   if type(popup["MDKPAttendance"]) ~= "table" then
      popup["MDKPAttendance"] = {
	 text = "",
	 button1 = TEXT(YES),
	 button2 = TEXT(NO),
	 timeout = 0,
	 whileDead = 1,
	 hideOnEscape = 0,
	 OnAccept = function() self:TakeAttendance() end
      }
   end
   
   MagicComm:RegisterListener(self, "MD")
   self:RegisterEvent("RAID_ROSTER_UPDATE", "StartOrEndRaid")
   self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
   self:ZONE_CHANGED_NEW_AREA()
   self:RegisterEvent("GUILD_ROSTER_UPDATE")
   GuildRoster()
end

function mod:ParseAltNote(note)
   if not note then return end
   local alt =  string.match(note, "Alt:(.*)")
   if alt then return strtrim(alt) end
end

local badData = { ['<AFK>'] = true, ['<DND>'] = true }
function mod:UpdatePlayerInfo(name, rank, level, class, race, sex, note, officernote)
   if badData[sex] then sex = nil end -- wtf
   if badData[race] then race = nil end -- ugly hack
   local hash = players[name] or {}
   hash.rank = rank or hash.rank or "Unknown"
   hash.level = level or hash.level or 0
   hash.class = class or hash.class or "Unknown"
   hash.race = race or hash.race or "Unknown"
   hash.sex = sex or hash.sex or 0
   players[name] = hash
   if db.detectAlts then
      local alt
      if note then
	 alt = self:ParseAltNote(note)
      end
      if not alt and officernote then
	 alt = self:ParseAltNote(officernote)
      end
      if alt then
	 if hash.main ~= alt then
	    if not players[alt] then
	       if self.hasWarn then self:warn("Found main character %s for %s, but main character not found.", alt, name) end
	    else
	       if self.hasInfo then self:info("Found main character %s for %s.", alt, name) end	       
	       hash.main = alt
	    end
	 end
	 if players[alt] then
	    if players[alt].main then
	       if self.hasWarn then self:warn("Removing main character %s from detected alt %s", players[alt].main, alt) end
	       players[alt].main = nil
	       mod:AddPlayerConfig(alt)
	    end
	 else
	    if self.hasSpam then self:spam("Detected main character %s (alt of %s) not found in player database.", alt, name) end
	 end
      elseif db.clearAlt and hash.main then
	 if self.hasWarn then self:warn("Removing main character %s from %s due to missing guild alt note.", hash.main, name) end
	 hash.main = nil
	 mod:AddPlayerConfig(name)
      end
   end
   mod:AddPlayerConfig(name)
   mod:NotifyChange()
end

function mod:GUILD_ROSTER_UPDATE()
   local offline = GetGuildRosterShowOffline()
   local selection = GetGuildRosterSelection()

   self:UnregisterEvent("GUILD_ROSTER_UPDATE") 
   SetGuildRosterShowOffline(true)
   SetGuildRosterSelection(0)
   GetGuildRosterInfo(0)
   local numGuildMembers = GetNumGuildMembers()
   for i = 1, numGuildMembers do
      local name, rank, rankIndex, level, class, zone, note, officernote, _, _, class = GetGuildRosterInfo(i)
      self:UpdatePlayerInfo(name, rank, level, class, nil, nil, note, officernote)
   end
   SetGuildRosterShowOffline(offline)
   SetGuildRosterSelection(selection)
end

function mod:TakeStandbyAttendance(event)
   local n = GetNumGuildMembers()
   if not n or n < 1 then return end

   self:UnregisterEvent("GUILD_ROSTER_UPDATE") 

   local offline = GetGuildRosterShowOffline()
   local selection = GetGuildRosterSelection()

   SetGuildRosterShowOffline(false)
   SetGuildRosterSelection(0)
   GetGuildRosterInfo(0)

   if db.reportstandby then
      mod:SendEventNotice(event.name, "GUILD")
   end

   local numGuildMembers = GetNumGuildMembers()
   for i = 1, numGuildMembers do
      local name, rank, rankIndex, level, class, zone, note, officernote, _, _, class = GetGuildRosterInfo(i)
      if not event.attendees[name] and not event.standby[name] and
	 (not db.ranks or db.ranks[rankIndex+1]) then
	 self:UpdatePlayerInfo(name, rank, level, class, nil, nil, note, officernote)
	 SetNetworkData("STANDBYCHECK", event.name, #raids)
	 self:SendUrgentMessage("WHISPER", name)
	 if self.hasTrace then self:trace("Sending attendance message to %s.", name) end
      end
   end
   SetGuildRosterShowOffline(offline)
   SetGuildRosterSelection(selection)
end


function mod:OnStandbyResponse(player, eventName, raidid)
   local raid
   if raidid and raids[raidid] then
      raid = raids[raidid]
   else
      raid = raids[#raids]
   end
   local event = raid.events[eventName]
   if event then
      if not event.attendees[player] then
	 event.standby[player] = true
	 if self.hasInfo then self:info("Adding %s as standby for event %s.", player, eventName) end
      else
	 if self.hasTrace then self:trace("Not adding %s to standby - already attending.", player) end
      end
   end
end

function mod:OnDisable()
   self:UnregisterEvent("CHAT_MSG_LOOT")
   self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
   self:UnregisterEvent("GUILD_ROSTER")
   self:UnregisterEvent("PLAYER_REGEN_DISABLED")
   self:UnregisterEvent("PLAYER_TARGET_CHANGED")
   self:UnregisterEvent("RAID_ROSTER_UPDATE")
   self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")   
   self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
   MagicComm:UnregisterListener(self, "MD")
end

local function UnitIsEligable (unit, uid)
   return UnitClassification(unit) == "worldboss" and (UnitCanAttack("player", unit)  or uid == 24892) -- special case for sathorarr since he's considered friendly when dead
end


local function GUIDToUID(guid)
   local uid = tonumber(sub(guid, 7, 12), 16)
   if uid == 0 then return end
   return uid
end

local function GetUnitID(unit)
   local guid = UnitGUID(unit)
   return guid, GUIDToUID(guid)
end


function mod:ScanTarget(unit)
   if not UnitExists(unit) then return end
   local guid, uid = GetUnitID(unit)

   if not currentZone or not UnitIsEligable(unit, uid) then
      return
   end

   if uid and not currentZone.mobs[uid] then
      self:AddNewBoss(uid, UnitName(unit), currentZone)
   end
   
   if UnitIsDead(unit) then
      -- we found a dead boss, check to see if this guy is killed
      if not currentEvent and currentRaid then
	 local event = currentZone.mobs[uid].event or currentZone.mobs[uid].name
	 local tmpEvent = currentRaid.events[event] 
	 if tmpEvent then
	    if self.hasTrace then self:trace("Found dead boss mob with finished event - activating.") end
	    currentEvent = tmpEvent
	    self:ScheduleTimer("ClearEvent", 120, currentEvent) -- clear it after 120 seconds
	 elseif not currentZone.mobs[uid].ignore then
	    if self.hasTrace then self:trace("Possibly logging event for [mouseover] %s: %s.", event, tostring(uid)) end
	    self:LogRaidEvent(uid, currentZone.mobs[uid])
	 end
      end
   elseif not loggingDeaths then
      if self.hasTrace then self:trace("Detected boss mob, enabling combat event log filtering.") end 
      self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED") -- eligable mob found, enable death tracking
      loggingDeaths = true
   end
end

-- From when friendly mobs were added
local blacklist = {
   24895, -- Madrigosa / SWP
   17948, -- Tyrande / Hyjal
   17773, -- Jaina / Hyjal
   17852, -- Thrall / Hyjal
   24891, -- Human Kalecgos, SWP
   23197, -- Maiev Shadowsong, BT
}
   
function mod:AddDefaultZones()
   if not mod.defaultZones then
      return
   end
   local key
   for zone, zonedata in pairs(mod.defaultZones) do
      key = self:MakeKey(zone)
      if not zones[key] then
	 zones[key] = {
	    name = zone,
	    size = zonedata.size,
	    mobs = { }
	 }
      end
      zone = zones[key].mobs
      for _, mob in ipairs(blacklist) do
	 zone[mob]  = nil -- remove these 
      end
      for mob, data in pairs(zonedata.mobs) do
	 if not zone[mob] then
	    zone[mob] = {}
	    for key,val in pairs(data) do
	       zone[mob][key] = val
	    end
	 end
      end
   end
end

function mod:ZONE_CHANGED_NEW_AREA()
   local zone,name = self:GetZoneName()
   currentZone = nil
   if zone == nil or zone == "" then
      self:ScheduleTimer("ZONE_CHANGED_NEW_AREA",5)
   else
      local inInstance, type = IsInInstance()
	 
      if type ~= "raid" then  -- FIXME
	 zone = "WorldBoss"
	 name = "World Boss"
      end
      if not zones[zone] then
	 zones[zone] = {
	    name = name, 
	    mobs = {}
	 }
	 self:BuildConfiguration()
	 self:NotifyChange()
      end

      currentZone = zones[zone]
      if not currentRaid then
	 self:StartOrEndRaid() -- make sure we have a raid if we should
      end
      if currentRaid and not currentRaid.zone and type == "raid" then
	 currentRaid.zone = name
	 self:NotifyChange()
	 if self.hasTrace then self:trace("Set %s to the zone of the current raid.", name) end
      end
   end
end

local linkpat = '|c........|Hitem:(%%d+):.-|h|r'
local globalpatterns = {
   "LOOT_ITEM",
   "LOOT_ITEM_SELF",
   "LOOT_ITEM_MULTIPLE",
   "LOOT_ITEM_SELF_MULTIPLE",
}

local patterns = {}
for _, global in ipairs(globalpatterns) do
   local pattern = _G[global]
   if sub(pattern, 1,2) == "%s" then
      pattern = "([^ ]+)"..sub(pattern, 3)
   end
   patterns[global] = gsub(gsub(pattern, "%%d", "(%%d+)"), "%%s", linkpat)
end
local npatterns = #patterns

function mod:CHAT_MSG_LOOT(event, msg)
   local looter, itemid, count

   if self.hasSpam then self:spam("Parsing loot event: %s", msg) end
   looter, itemid, count = match(msg, patterns["LOOT_ITEM_MULTIPLE"])

   if looter and itemid and count then
      mod:HandleLoot(looter, tonumber(itemid), tonumber(count))
      return
   end
   
   itemid, count  = match(msg, patterns["LOOT_ITEM_SELF_MULTIPLE"])
   if itemid and count then
      mod:HandleLoot(playerName, tonumber(itemid), tonumber(count))
      return
   end

   looter, itemid = match(msg, patterns["LOOT_ITEM"])
   if looter and itemid then
      mod:HandleLoot(looter, tonumber(itemid), 1)
      return
   end

   itemid = match(msg, patterns["LOOT_ITEM_SELF"])
   if itemid then
      mod:HandleLoot(playerName, tonumber(itemid), 1)
      return
   end
end

function mod:AddItemManually(_,cmd)
   if not db.enabled then
      if self.error then self:error("The addon isn't enabled.") end
      return
   end
   local looter,itemid = match(cmd, "([^ ]+) .*|Hitem:(%d+):")
   if not looter or not itemid then
      if self.error then self:error("Invalid syntax. Expected: <looter> [itemlink], got '%s'", cmd) end
      return
   end
   self:HandleLoot(looter, tonumber(itemid), 1, true)
end

function mod:HandleLoot(looter, itemid, count, force)
   if not currentRaid then
      self:Print("Unable to add item - no active raid found.")
      return
   end
   local name,link,rarity = GetItemInfo(itemid)
   if self.hasSpam then self:spam("handling %s", link) end

   if not force and rarity < 4 and not items[itemid] then
      if self.hasSpam then self:spam("Ignoring %s [low item quality]", link) end
      return
   end

   if not items[itemid] then
      self:UpdateItem(itemid)
   end

   local itemData = items[itemid]

   if itemData.ignore and not force then
      if self.hasSpam then self:spam("Ignoring %s [ignored in database]", link) end
      return
   end
   if self.hasSpam then self:spam("adding %s ", link) end

   if itemData.bank then
      self:AddItem("bank", itemid, 0, count)
   elseif itemData.fixed or db.trackonly then
      self:AddItem(looter, itemid, (itemData.value or 0)*count, count)
   else 
      QueueItem(looter, itemid, (itemData.value or 0), link, count)
   end
end

function mod:SendEventNotice(name, dest)
   if dest == "RAID" then
      SendChatMessage(L["MagicDKP: Logging raid attendance for "]..name, dest)
   else
      SendChatMessage(L["MagicDKP: Taking standby DKP for event "]..name, dest)
   end
end

      
function mod:TakeAttendance()
   local groups = 5
   if currentZone and currentZone.size then
      groups = math.floor(currentZone.size / 5)
   end
   if not currentEvent then return end
   currentEvent.eventTime = self:GetGameTimeAsUnix()

   if db.reportraid then
      mod:SendEventNotice(currentEvent.name, "RAID")
   end
   
   for i = 1, GetNumRaidMembers() do  
      local name, _, group, level, _, class = GetRaidRosterInfo(i)
      local id = "raid"..i
      self:UpdatePlayerInfo(name, nil, level == 0 and nil or level, class, select(2, UnitRace(id)), UnitSex(id))
      if group <= groups then
	 currentEvent.attendees[name] = true
	 if self.hasTrace then self:trace("Adding %s to %s.", name, currentEvent.name) end
      else
	 currentEvent.standby[name] = true
	 if self.hasTrace then self:trace("Adding %s as standby to %s.", name, currentEvent.name) end
      end
   end

   self:RegisterEvent("GUILD_ROSTER_UPDATE", "TakeStandbyAttendance", currentEvent)
   GuildRoster()
   
   currentRaid.events[currentEvent.name] = currentEvent
   self:BuildConfiguration()
   self:ScheduleTimer("ClearEvent", 300, currentEvent)
   self:NotifyChange()
end


local raidAttendanceText = L["Record attendance for %s?"]
local eventInProgress

function mod:LogRaidEvent(uid, unitData)
   if not currentRaid then return end -- should never happen
   local event = unitData.event or unitData.name 
   if not eventInProgress or event ~= eventInProgress.name then
      if currentRaid.events[event] then
	 eventInProgress = currentRaid.events[event]
      else
	 if self.hasTrace then self:trace("Creating a new event...") end
	 eventInProgress = {
	    attendees = {},
	    standby = {},
	    items = {},
	    needed = {},
	    name = event,
	    zone = currentZone.name,
	 }
	 for mobuid,data in pairs(currentZone.mobs) do
	    if (data.event or data.name) == event and not data.ignore then
	       eventInProgress.needed[mobuid] = true
	    end
	 end
      end
   end

   eventInProgress.needed[uid] = nil

   if not next(eventInProgress.needed) then 
      eventInProgress.needed = nil -- don't need to save this
      if self.hasTrace then self:trace("Event finished, taking attendance...") end
      if loggingDeaths then
	 loggingDeaths = false
	 -- we're done with this event so no longer need track this
	 self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED") 
      end
      currentEvent = eventInProgress
      eventInProgress = nil
      currentBoss = nil
      if not currentEvent.eventTime then
	 StaticPopupDialogs["MDKPAttendance"].text = raidAttendanceText:format(event)
	 StaticPopup_Show("MDKPAttendance")
      end
   end
end


function mod:ClearEvent(event)
   if event == currentEvent then
      currentEvent = nil
   end
end
function mod:PLAYER_REGEN_DISABLED()
   eventInProgress = nil
   currentBoss = nil
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, _, _, guid)
   local uid
   if not currentZone or not currentRaid then
      self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
      return
   end
   if not currentBoss then
      uid = GUIDToUID(guid) 
      if currentZone.mobs[uid] and not currentZone.mobs[uid].isfarm  then
	 -- This means we're engaged with a boss that gives wipe dkp
	 if self.hasTrace then self:trace("Engaged boss: %s", currentZone.mobs[uid].name) end
	 currentBoss = uid
      end
   end
   
   if event ~= "UNIT_DIED" and event ~= "PARTY_KILL" then return end

   uid = uid or GUIDToUID(guid)
   
   if uid and currentZone.mobs[uid] and not currentZone.mobs[uid].ignore then
      if self.hasTrace then self:trace("Possibly logging event for %s: %s.", event, tostring(uid)) end
      self:LogRaidEvent(uid, currentZone.mobs[uid])
   end
end

local isInRaid

-- Get the current raid or create a new one if none is active
function mod:FindOrCreateRaid()
   local raid = raids[#raids]
   local changed = false
   if not raid or raid.endTime then
      raid = {
	 startTime = self:GetGameTimeAsUnix(),
	 events = {
	 },
      }
      raids[#raids+1] = raid
      if self.hasDebug then self:debug("Created new raid.") end
      changed = true
      self:AddRaidConfig(#raids)
   else
      if self.hasDebug then self:debug("Found active raid.") end
   end
   currentRaid = raid
   if not currentZone then
      self:ZONE_CHANGED_NEW_AREA()
   end
   if not raid.zone and currentZone.name ~= "World Boss" then
      raid.zone = currentZone.name
      changed = true
   end
   if changed then
      self:NotifyChange()
   end
end

function mod:EndRaid()
   currentRaid = raids[#raids]
   if currentRaid and not currentRaid.endTime  then
      if next(currentRaid.events) then
	 currentRaid.endTime = self:GetGameTimeAsUnix()
	 if self.hasDebug then self:debug("Ending active raid.") end
	 self:NotifyChange()
      else
	 -- No events, don't save
	 self:DeleteRaidConfig(#raids)
	 raids[#raids] = nil
	 if self.hasDebug then self:debug("Deleting active raid (no events).") end
	 self:NotifyChange()
      end
   elseif self.hasTrace then
      self:trace("No active raid found.")
   end
   currentRaid = nil
   currentZone = nil
end

function mod:StartOrEndRaid()
   local num = GetNumRaidMembers()
   if currentRaid and  currentRaid.endTime then
      currentRaid = nil
   end
   if num >= tonumber(db.raidsize) then
      if not currentRaid then
	 if self.hasTrace then self:trace("No current raid - creating new or finding active. "); end
	 self:FindOrCreateRaid()
      end
      self:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "ScanTarget", "mouseover")   
      self:RegisterEvent("PLAYER_TARGET_CHANGED", "ScanTarget", "target")   
      self:RegisterEvent("CHAT_MSG_LOOT")	    
      self:RegisterEvent("PLAYER_REGEN_DISABLED")
   elseif num == 0 then
      if not currentRaid then
	 currentRaid = raids[#raids]
      end
      if currentRaid and not currentRaid.endTime then
	 if self.hasTrace then self:trace("Current raid should be ended.") end
	 self:EndRaid() -- we're no longer in a raid at all, so end the raid
      end
      self:UnregisterEvent("PLAYER_REGEN_DISABLED")
      self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")   
      self:UnregisterEvent("PLAYER_TARGET_CHANGED")
      self:UnregisterEvent("CHAT_MSG_LOOT")
   end   
end

function mod:GetGameTimeAsUnix()
   -- This does NOT work - if local time just swapped across midnight, but gametime is nto we get a 24 hr error.
   --   local h, m = GetGameTime()
   --   local n = date("*t")
   --   n.hour = h
   --   n.min = m
   --   return time(n)
   -- 
   return time() 
end

function mod:FromUnixTime(timestamp, prefix, postfix, short)
   local dt = date("*t", timestamp)
   if short then
      return string.format("%s%02d/%02d %s", prefix or "" , dt.month, dt.day, postfix or "")
   else
      return string.format("%s%02d/%02d/%d %02d:%02d:%02d%s", prefix or "" , dt.month, dt.day, dt.year, dt.hour, dt.min, dt.sec, postfix or "")
   end
end

function mod:LogCustomEvent(eventid)
   if not currentRaid then
      if self.error then self:error("No active raid - not logging event '%s'", events[eventid]) end
      return
   end
   local oldevent = currentEvent
   currentEvent =  currentRaid.events[events[eventid]]
   if not currentEvent then
      currentEvent = {
	 attendees = {},
	 standby = {},
	 name = events[eventid],
	 zone = currentZone.name,
      }
   end

   if not currentEvent.eventTime or not next(currentEvent.attendees) then
      if self.hasInfo then self:info("Taking Attendance for: %s", events[eventid]) end
      self:TakeAttendance()
   elseif self.hasInfo then
      self:info("Event %s already logged with attendance.", events[eventid])
   end
   currentEvent = oldevent
end

function mod:AddWipe()
   if not currentRaid then
      if self.error then self:error("No active raid - not logging wipe.") end
      return
   elseif not currentBoss then
      if self.error then self:error("No boss recently engaged - not logging wipe.") end
      return
   end

   
   
   local event = string.format("%s - %s #", L["Wipe"],
			       currentZone.mobs[currentBoss].event or
				  BL[currentZone.mobs[currentBoss].name] or
				  currentZone.mobs[currentBoss].name)

   local eventnum
   currentBoss = nil
   eventInProgress = nil
   for id = 1, 100 do
      eventnum = event..id
      if not currentRaid.events[eventnum] then
	 currentEvent = {
	    attendees = {},
	    standby = {},
	    zone = currentZone.name,
	    name = eventnum,
	 }
	 currentRaid.events[eventnum] = currentEvent
	 if self.hasInfo then self:info("Logging Wipe: %s", eventnum) end
	 self:TakeAttendance()
	 currentEvent = nil
	 return
      end
   end
end

function mod:OnProfileChanged(event, newdb)
   if event ~= "OnProfileDeleted" then
      db = self.db.profile
      self:SetLogLevel(db.logLevel)
   end
      
   self:NotifyChange()

end


local exportDialog


-- Ugly export function for an ugly XML format, written in an ugly way
-- to work with a stupid importer. Yeah this is some messy stuff.
function mod:ExportCTRT(info)
   local itempat = '|c(........)|Hitem:(.-)|h%['
   local raidid = tonumber(type(info) == "table" and info[#info-1] or info)
   if raids[raidid] then
      local dummyItemId = 21877 -- netherweave
      if not GetItemInfo(dummyItemId) then
	 dummyItemId = 33470 -- frostweave
      end
      if not GetItemInfo(dummyItemId) then
	 dummyItemId = 33447 -- runic healing potion
      end
      local itemdummy = {
	 {
	    value = 0,
	    looter = "disenchant",
	    itemid = dummyItemId,
	    count = 1
	 }
      }
      local raid = raids[raidid]
      local loot = ""
      local lootcount, count = 1, 1
      local zone = raid.zone or "Unknown"
      local export = "<RaidInfo>"
	 .. self:FromUnixTime(raid.startTime, "<key>", "</key>")
	 .. self:FromUnixTime(raid.startTime, "<start>", "</start>")
	 .. self:FromUnixTime(raid.endTime or raid.startTime , "<end>", "</end>")
	 .. "<zone>"..zone.."</zone>"

      local bosskills = ""

      local jointxt, leavetxt = "", ""
      local joincount = 1
      local sortevents = {}
      for id in pairs(raid.events) do
	 sortevents[#sortevents+1] = id
      end
      table.sort(sortevents,
		 function(a, b)
		    return (raid.events[a].eventTime or 0) < (raid.events[b].eventTime or 0)
		 end)
      local alts = {}
      for id,data in pairs(players) do
	 if data.main then
	    alts[id] = data.main
	 end
      end
      local addevent = function(data, id)
			  local attendees = ""
			  local acount = 1
			  local eventTime = data.eventTime or raid.startTime
			  local eventTimeStr = self:FromUnixTime(eventTime)
			  
			  local attendee = function(name)
					      local pinfo
					      name = alts[name] or name
					      pinfo = players[name]
					      attendees = string.format("%s<key%d><name>%s</name></key%d>", attendees, acount, name, acount)
					      if pinfo then
						 if pinfo.sex and badData[pinfo.sex] then pinfo.sex = nil end
						 if pinfo.race and badData[pinfo.race] then pinfo.race = nil end
					      end
					      jointxt = string.format("%s<key%d><player>%s</player><race>%s</race><class>%s</class><sex>%d</sex><level>%d</level><time>%s</time></key%d>", jointxt, joincount, name,
								      pinfo and pinfo.race or "", pinfo and pinfo.class or "", tonumber(pinfo and pinfo.sex) or 0, pinfo and pinfo.level or 0, 
								      self:FromUnixTime(eventTime-1), joincount)
					      leavetxt = string.format("%s<key%d><player>%s</player><time>%s</time></key%d>", leavetxt, joincount, name, self:FromUnixTime(eventTime+1), joincount)
					      acount = acount + 1
					      joincount = joincount + 1
					   end
			  
			  for name in pairs(data.attendees) do
			     attendee(name)
			  end
			  
			  if db.standbyattendee and data.standby then
			     for name in pairs(data.standby) do
				attendee(name)
			     end
			  end
			  
			  bosskills = string.format("%s<key%d><name>%s</name><time>%s</time><attendees>%s</attendees></key%d>",
						    bosskills, count, id, eventTimeStr, attendees, count)
			  count = count + 1
			  local tmpitems
			  local overridezone
			  if data.items and next(data.items) then
			     tmpitems = data.items
			  else
			     tmpitems = itemdummy -- Stupid CTRT wont' import event without items
			     if db.eventnameaszone then
				for _, data in pairs(events) do
				   if id == data then
				      overridezone = id
				      break
				   end
				end
			     end
			  end
			  
			  for _,item in pairs(tmpitems) do
			     local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
			     itemSubType, itemStackCount, itemEquipLoc, itemTexture
			     if items[item.itemid] then
				itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
				itemSubType, itemStackCount, itemEquipLoc, itemTexture =  unpack(items[item.itemid].itemInfo)
			     end
			     
			     if not itemName then 
				itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
				itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item.itemid)
			     end
			     
			     --|cffffffff|Hitem:23984:0:0:0:0:0:0:0|h[Irradiated Crystal Shard]|h|
			     if self.hasSpam then self:spam("Exporting item %d : %s => %s", item.itemid, tostring(itemName), tostring(itemLink)) end
			     
			     local color, link = match(itemLink, itempat)
			     loot = string.format("%s<key%d><ItemName>%s</ItemName><ItemID>%s</ItemID><Icon>%s</Icon><Class>%s</Class><SubClass>%s</SubClass><Color>%s</Color><Count>%d</Count><Player>%s</Player><Costs>%d</Costs><Time>%s</Time><Zone>%s</Zone><Boss>%s</Boss><Note><![CDATA[ - Zone: %s - Boss: %s - %d DKP]]></Note></key%d>",
						  loot, lootcount, itemName, link, itemTexture, itemType, itemSubType,color,
						  item.count, alts[item.looter] or item.looter, item.value or 0, eventTimeStr,
						  overridezone or data.zone or zone, id, overridezone or data.zone or zone,
						  id, item.value or 0, lootcount)
			     lootcount = lootcount + 1
			  end
		       end
      local cache = {}
      for _,id in pairs(sortevents) do
	 local data = raid.events[id]
	 addevent(data, id)
      end
      export = string.format("%s<BossKills>%s</BossKills><Join>%s</Join><Leave>%s</Leave><Loot>%s</Loot></RaidInfo>", export, bosskills, jointxt, leavetxt, loot)
      if not exportDialog then
	 exportDialog = GUI:Create("MDKP_Dialog")
	 exportDialog:SetTitle(L["CTRT Export"])
	 exportDialog:SetLayout("Fill")
	 exportDialog.editbox:SetWidth(350)
	 exportDialog.editbox:SetNumeric(false)
	 exportDialog:SetText(L["Copy the data below to export in CTRT format."])

	 exportDialog.close = GUI:Create("Button")
	       
	 exportDialog:AddChild(exportDialog.close)
	 exportDialog.close:SetText(L["Close"])
	 
	 exportDialog.close:SetCallback("OnClick",
					function()
					   exportDialog:Hide()
					   exportDialog.editbox:SetText("")
					end
				     )
	 
      end
      exportDialog.editbox:SetText(export)
      exportDialog:Show()
   end
end

function mod:SendUrgentMessage(channel, recipient)
   MagicComm:SendUrgentMessage(networkData, "MD", channel, recipient)
end

function mod:SendBulkMessage()
   if MagicMarker:IsValidMarker() then
      MagicComm:SendBulkMessage(networkData, "MD")
   end
end

do
   local verRespMsg = "%s: %s revision %s"
   function mod:OnVersionResponse(ver, major, minor, sender)
      self:Print(verRespMsg:format(sender, major or "Unknown", minor or "Unknown"))
   end
end

function mod:QueryAddonVersions()
   SetNetworkData("VCHECK")
   self:SendUrgentMessage("GUILD")
   self:SendUrgentMessage("RAID")
end
